--------------------------------------------------------------------------------
-- SubModule GPSTransmitter
-- Created   12/19/2007 6:15:17 PM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity GPSTransmitter is port
   (
     CLK       : in    std_logic;
     RST       : in    std_logic;
     PPS       : in    std_logic;
     INIT      : in    std_logic;
     HALT      : in    std_logic;
     POSHOLD   : in    std_logic;
     UNHOLD    : in    std_logic;

     D         : out   std_logic_vector(7 downto 0);
     LD        : out   std_logic;
     BUSY      : in    std_logic
   );
end GPSTransmitter;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of GPSTransmitter is

-- Constant Declarations
   constant ini : string    := "@@Aw" & nul & '6' & cr & lf &-- set GPS time
                               "@@Ha" & soh & '(' & cr & lf; -- turn on Ha once a sec
--                             "@@Ge" & soh * '#' & cr & lf; -- turn on T-RAIM
   constant rep : string    := "@@Bj" & nul & '(' & cr & lf; -- query leap pending
   constant stop : string   := "@@Hb" & nul & '*' & cr & lf; -- turn off Hb
   constant survey : string := "@@Gd" & etx & ' ' & cr & lf; -- set auto-survey
   constant pos3d : string  := "@@Gd" & nul & '#' & cr & lf; -- back to 3d positioning
   constant max_len : natural := ini'length;

-- Type Declaration
   type FSM_state is (IDLE, SEND, WAITING);

-- Signal Declarations
   signal cnt                : unsigned (11 downto 0);
   signal endcnt             : unsigned (11 downto 0);
   signal state              : FSM_state;
--   signal nextstate          : FSM_state;
   signal pendreq            : std_logic;
   signal reqcnt             : unsigned (3 downto 0) := (others => '0');
   signal haltcnt            : unsigned (3 downto 0) := (others => '0');
   signal oldpps             : std_logic;

   signal posrst             : std_logic;
   signal start_survey       : std_logic;
   signal start_3d           : std_logic;

-- Functions
   function to_std_logic_vector(a : string) return std_logic_vector is
      variable ret : std_logic_vector(a'length*8-1 downto 0);
   begin
      for i in a'range loop
         ret(i*8-1 downto (i-1)*8) := std_logic_vector(to_unsigned(character'pos(a(i)), 8));
      end loop;
      return ret;
   end function to_std_logic_vector;

   function get_byte (s : in string; i : positive) return std_logic_vector is
      variable c : character;
   begin
      c := s (i);
      return std_logic_vector (To_unsigned (character'pos(c), 8));
   end function get_byte;

begin
   -- latch position hold and unhold requests
   process (CLK) is
      variable TMP: std_logic := '0';
   begin
      if rising_edge (CLK) then
         if (posrst = '1') then
            TMP := '0';
         elsif (POSHOLD = '1') then
            TMP := '1';
         else
            TMP := TMP;
         end if;
      end if;
      start_survey <= TMP;
   end process;
   process (CLK) is
      variable TMP: std_logic := '0';
   begin
      if rising_edge (CLK) then
         if (posrst = '1') then
            TMP := '0';
         elsif (UNHOLD = '1') then
            TMP := '1';
         else
            TMP := TMP;
         end if;
      end if;
      start_3d <= TMP;
   end process;

   -- repeat requests every 16s
   process (CLK) is
   begin
      if rising_edge (CLK) then
         oldpps <= pps;
         pendreq <= '0';
         if (oldpps = '0') and (pps = '1') then
            reqcnt <= reqcnt + 1;
            if reqcnt = to_unsigned (7, 4) then
               pendreq <= '1';
            end if;
         end if;
      end if;
   end process;

   -- main state machine
   process (CLK) is
      variable s : std_logic_vector (8*max_len-1 downto 0);
   begin
      if rising_edge (CLK) then
         if RST = '1' then
             cnt <= To_unsigned (1, 12);
             endcnt <= To_unsigned (ini'length, 12);
             s(8*ini'length-1 downto 0) := to_std_logic_vector (ini);
             D <= (others => '0');
             LD <= '0';
             posrst <= '1';
             state <= SEND;
         else
             posrst <= '0';
             LD <= '0';
             case state is
                when IDLE =>
                   cnt <= To_unsigned (1, 12);
                   -- send intialization message
                   if INIT = '1' then
                      endcnt <= To_unsigned (ini'length, 12);
                      s(8*ini'length-1 downto 0) := to_std_logic_vector (ini);
                      state <= SEND;
                   elsif HALT = '1' then
                      haltcnt <= haltcnt + 1;
                      if (haltcnt = to_unsigned (7, 4)) then
                         -- send stop message for Hb
                         endcnt <= To_unsigned (stop'length, 12);
                         s(8*stop'length-1 downto 0) := to_std_logic_vector (stop);
                         state <= SEND;
                      end if;
                   -- Request leap second update
                   elsif pendreq = '1' then
                      endcnt <= To_unsigned (rep'length, 12);
                      s(8*rep'length-1 downto 0) := to_std_logic_vector (rep);
                      state <= SEND;
                   -- Request position hold survey
                   elsif start_survey = '1' then
                      posrst <= '1';
                      endcnt <= To_unsigned (survey'length, 12);
                      s(8*survey'length-1 downto 0) := to_std_logic_vector (survey);
                      state <= SEND;
                   -- Request 3d position
                   elsif start_3d = '1' then
                      posrst <= '1';
                      endcnt <= To_unsigned (pos3d'length, 12);
                      s(8*pos3d'length-1 downto 0) := to_std_logic_vector (pos3d);
                      state <= SEND;
                   end if;
                when SEND =>
--                   nextstate <= SEND;
                   if BUSY = '0' then
                      -- D <= get_byte (s, To_integer (cnt));
                      D <= s (7 downto 0);
                      s := X"00" & s (s'length-1 downto 8);
                      LD <= '1';
                      cnt <= cnt + 1;
                      if cnt = endcnt then
                         state <= IDLE;
                      else
                         state <= WAITING;
                      end if;
                   end if;
--                when SENDSTOP =>
--                   nextstate <= SENDSTOP;
--                   if BUSY = '0' then
--                      D <= get_byte (stop, To_integer (cnt)+1);
--                      LD <= '1';
--                      cnt <= cnt + 1;
--                      if cnt = To_unsigned (stop_len - 1, 12) then
--                         state <= IDLE;
--                      else
--                         state <= WAITING;
--                      end if;
--                   end if;
--                when SENDREP =>
--                   nextstate <= SENDREP;
--                   if BUSY = '0' then
--                      D <= get_byte (rep, To_integer (cnt)+1);
--                      LD <= '1';
--                      cnt <= cnt + 1;
--                      if cnt = To_unsigned (rep_len - 1, 12) then
--                         state <= IDLE;
--                      else
--                         state <= WAITING;
--                      end if;
--                   end if;
                when WAITING =>
                  state <= SEND;
             end case;
          end if;
       end if;
   end process;

end Structure;
--------------------------------------------------------------------------------
