--------------------------------------------------------------------------------
-- SubModule TimingLinkWrap
-- Created   8/26/2008 10:29:10 AM
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity TimingLinkWrap is port
   (
     CLK            : in    std_logic;
     RST            : in    std_logic;
     START          : in    std_logic;
     ENABLE         : in    std_logic;
     RDY            : out   std_logic;
     DONE           : out   std_logic;
     DOUT           : out   std_logic_vector(7 downto 0);
     PREV_START     : out   std_logic;
     PREV_ENABLE    : out   std_logic;
     PREV_RDY       : in    std_logic;
     PREV_DONE      : in    std_logic;
     PREV_DIN       : in    std_logic_vector(7 downto 0)
   );
end TimingLinkWrap;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TimingLinkWrap is

-- Constant Declarations
   constant header : string := "LIGO TIMING SYSTEM VERSION 1.0" & cr & lf;
   constant header_len : positive := header'length(1);

-- Constant Declarations
   constant CRC32_POLYNOMIAL : std_logic_vector (31 downto 0) := X"04C11DB7";
   constant CRC32_INIT_VALUE : std_logic_vector (31 downto 0) := X"FFFFFFFF";
   constant CRC32_MATCH_VALUE: std_logic_vector (31 downto 0) := X"C704DD7B";

-- Component Declarations
   component ucrc_par
      generic (
         POLYNOMIAL : std_logic_vector;
         INIT_VALUE : std_logic_vector;
         DATA_WIDTH : integer range 2 to 256;
         SYNC_RESET : integer range 0 to 1;
         MATCH_VALUE : std_logic_vector);  -- use synchronous reset
      port (
         clk_i   : in  std_logic;       -- clock
         rst_i   : in  std_logic;       -- init CRC
         clken_i : in  std_logic;       -- clock enable
         data_i  : in  std_logic_vector(DATA_WIDTH - 1 downto 0);  -- data input
         match_o : out std_logic;       -- CRC match flag
         crc_o   : out std_logic_vector(POLYNOMIAL'length - 1 downto 0));  -- CRC output
   end component;

-- Function Declarations
   function reverse_any_vector (a: in std_logic_vector)
         return std_logic_vector is
      variable result: std_logic_vector(a'REVERSE_RANGE);
   begin
      for i in a'RANGE loop
         result(i) := a(i);
      end loop;
      return result;
   end function reverse_any_vector;

-- Functions
   function get_byte (s : in string; i : positive) return std_logic_vector is
      variable c : character;
   begin
      c := s (i);
      return std_logic_vector (To_unsigned (character'pos(c), 8));
   end function get_byte;

-- Type Declarations
   type char_state_type  is (IDLE, SHIFT_HEADER_READ, SHIFT_HEADER_WRITE,
                             SHIFT_PREV_WAIT, SHIFT_PREV_READ, SHIFT_PREV_WRITE,
                             SHIFT_CRC_INIT, SHIFT_CRC_READ, SHIFT_CRC_WRITE,
                             SHIFT_TEST_READ, SHIFT_TEST_WRITE);

-- Signal Declarations
   signal char_dout      : std_logic_vector (7 downto 0);
   signal char_done      : std_logic;
   signal char_d         : std_logic_vector (7 downto 0);
   signal char_prev_en   : std_logic;
   signal char_rdy       : std_logic;
   signal char_rdy_old   : std_logic;
   signal char_rdy_old2  : std_logic;
   signal char_count     : unsigned (15 downto 0);
   signal char_state     : char_state_type;
   signal crc_val        : std_logic_vector (31 downto 0);
   signal crc_reg        : std_logic_vector (31 downto 0);
   signal crc_d          : std_logic_vector (7 downto 0);
   signal crc_rdy        : std_logic;

begin

   -- CRC generator
   char_crc: ucrc_par
      generic map (
         POLYNOMIAL => CRC32_POLYNOMIAL,
         INIT_VALUE => CRC32_INIT_VALUE,
         DATA_WIDTH => 8,
         SYNC_RESET => 0,
         MATCH_VALUE => CRC32_MATCH_VALUE)
      port map (
         clk_i   => CLK,
         rst_i   => START,
         clken_i => crc_rdy,
         data_i  => crc_d,
         match_o => open,
         crc_o   => crc_val
      );
   --crc_d <= reverse_any_vector (char_d);
   crc_d <= char_d;
   char_rdy_old <= char_rdy when rising_edge (CLK);
   char_rdy_old2 <= char_rdy_old when rising_edge (CLK);
   crc_rdy <= char_rdy_old and not char_rdy_old2;

   -- character interface state machine
   char_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            char_state <= IDLE;
            char_done <= '1';
            char_prev_en <= '0';
            char_rdy <= '0';
         elsif START = '1' then
            char_state <= SHIFT_HEADER_READ;
            char_count <= To_Unsigned (0, 16);
            char_done <= '0';
            char_prev_en <= '0';
            char_rdy <= '0';
         else
            char_prev_en <= '0';
            case char_state is
               when IDLE =>
                  -- nothing
                  char_rdy <= '0';
                  char_done <= '1';
                  char_count <= To_Unsigned (0, 16);

               -- shift out header but only if lastlink is true
               when SHIFT_HEADER_READ =>
                  char_d <= get_byte (header, To_integer (char_count));
                  char_count <= char_count + 1;
                  char_rdy <= '1';
                  char_state <= SHIFT_HEADER_WRITE;
               when SHIFT_HEADER_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     if char_count = To_Unsigned (header_len, 16) then
                        char_state <= SHIFT_PREV_WAIT;
                        --char_state <= SHIFT_TEST_READ;
                        char_count <= To_Unsigned (0, 16);
                     else
                        char_state <= SHIFT_HEADER_READ;
                     end if;
                  end if;

               -- test pattern
               when SHIFT_TEST_READ =>
                  char_d <= std_logic_vector (char_count (7 downto 0));
                  char_count <= char_count + 1;
                  char_rdy <= '1';
                  char_state <= SHIFT_TEST_WRITE;
               when SHIFT_TEST_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     if char_count = To_Unsigned (2304, 16) then
                        char_state <= SHIFT_CRC_INIT;
                        char_count <= To_Unsigned (0, 16);
                     else
                        char_state <= SHIFT_TEST_READ;
                     end if;
                  end if;

               -- shift out data from previous link
               when SHIFT_PREV_WAIT =>
                  char_rdy <= '0';
                  if PREV_DONE = '1' then
                     char_state <= SHIFT_CRC_INIT;
                  elsif PREV_RDY = '1' then
                     char_state <= SHIFT_PREV_READ;
                     char_prev_en <= '1';
                  end if;
               when SHIFT_PREV_READ =>
                  char_d <= PREV_DIN;
                  char_rdy <= '1';
                  char_state <= SHIFT_PREV_WRITE;
               when SHIFT_PREV_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     char_state <= SHIFT_PREV_WAIT;
                  end if;

               -- shift out CRC
               when SHIFT_CRC_INIT =>
                  crc_reg (31 downto 24) <= not reverse_any_vector (crc_val (31 downto 24));
                  crc_reg (23 downto 16) <= not reverse_any_vector (crc_val (23 downto 16));
                  crc_reg (15 downto  8) <= not reverse_any_vector (crc_val (15 downto  8));
                  crc_reg ( 7 downto  0) <= not reverse_any_vector (crc_val ( 7 downto  0));
                  -- crc_reg <= not reverse_any_vector (crc_val);
                  char_state <= SHIFT_CRC_READ;
                  char_count <= To_Unsigned (0, 16);
               when SHIFT_CRC_READ =>
                  char_d <= crc_reg (31 downto 24);
                  crc_reg <= crc_reg (23 downto 0) & X"00";
                  char_count <= char_count + 1;
                  char_rdy <= '1';
                  char_state <= SHIFT_CRC_WRITE;
               when SHIFT_CRC_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     if char_count = To_Unsigned (4, 16) then
                        char_done <= '1';
                        char_state <= IDLE;
                     else
                        char_state <= SHIFT_CRC_READ;
                     end if;
                  end if;
             end case;
         end if;
      end if;
   end process char_fsm;

   PREV_START <= START when rising_edge (CLK);
   PREV_ENABLE <= char_prev_en;
   RDY <= char_rdy;
   DONE <= char_done;
   DOUT <= char_d;

end Structure;
--------------------------------------------------------------------------------
